@@ -1,7 +1,7 @@
 /*
  * OsmoGGSN - Gateway GPRS Support Node
  * Copyright (C) 2002, 2003, 2004 Mondru AB.
- * Copyright (C) 2017 by Harald Welte <laforge@gnumonks.org>
+ * Copyright (C) 2017-2019 by Harald Welte <laforge@gnumonks.org>
  *
  * The contents of this file may be used under the terms of the GNU
  * General Public License Version 2, provided that the above copyright
@@ -42,21 +42,8 @@
 #include <netinet/ip.h>
 #include <netinet/ip6.h>
 
-#include <osmocom/core/application.h>
-#include <osmocom/core/select.h>
-#include <osmocom/core/stats.h>
-#include <osmocom/core/rate_ctr.h>
 #include <osmocom/core/timer.h>
 #include <osmocom/ctrl/control_if.h>
-#include <osmocom/ctrl/control_cmd.h>
-#include <osmocom/ctrl/control_vty.h>
-#include <osmocom/ctrl/ports.h>
-#include <osmocom/vty/telnet_interface.h>
-#include <osmocom/vty/logging.h>
-#include <osmocom/vty/stats.h>
-#include <osmocom/vty/ports.h>
-#include <osmocom/vty/command.h>
-#include <osmocom/vty/misc.h>
 #include <osmocom/gsm/apn.h>
 
 #include "../lib/tun.h"
@@ -64,31 +51,26 @@
 #include "../lib/syserr.h"
 #include "../lib/in46_addr.h"
 #include "../lib/gtp-kernel.h"
+#include "../lib/util.h"
 #include "../gtp/pdp.h"
 #include "../gtp/gtp.h"
-#include "icmpv6.h"
+#include "../lib/icmpv6.h"
+#include "pco.h"
 #include "ggsn.h"
 
-void *tall_ggsn_ctx;
-
-static int end = 0;
-static int daemonize = 0;
-static struct ctrl_handle *g_ctrlh;
-
-struct ul255_t qos;
-struct ul255_t apn;
-
-#define LOGPAPN(level, apn, fmt, args...)			\
-	LOGP(DGGSN, level, "APN(%s): " fmt, (apn)->cfg.name, ## args)
-
-#define LOGPGGSN(level, ggsn, fmt, args...)			\
-	LOGP(DGGSN, level, "GGSN(%s): " fmt, (ggsn)->cfg.name, ## args)
-
-#define LOGPPDP(level, pdp, fmt, args...) LOGPDPX(DGGSN, level, pdp, fmt, ## args)
-
 static int ggsn_tun_fd_cb(struct osmo_fd *fd, unsigned int what);
 static int cb_tun_ind(struct tun_t *tun, void *pack, unsigned len);
 
+void ggsn_close_one_pdp(struct pdp_t *pdp)
+{
+	LOGPPDP(LOGL_DEBUG, pdp, "Sending DELETE PDP CTX due to shutdown\n");
+	gtp_delete_context_req2(pdp->gsn, pdp, NULL, 1);
+	/* We have nothing more to do with pdp ctx, free it. Upon cb_delete_context
+	   called during this call we'll clean up ggsn related stuff attached to this
+	   pdp context. After this call, ippool member is cleared so
+	   data is no longer valid and should not be accessed anymore. */
+	gtp_freepdp_teardown(pdp->gsn, pdp);
+}
 
 static void pool_close_all_pdp(struct ippool_t *pool)
 {
@@ -106,14 +88,13 @@ static void pool_close_all_pdp(struct ippool_t *pool)
 		pdp = member->peer;
 		if (!pdp)
 			continue;
-		LOGPPDP(LOGL_DEBUG, pdp, "Sending DELETE PDP CTX due to shutdown\n");
-		gtp_delete_context_req(pdp->gsn, pdp, NULL, 1);
+		ggsn_close_one_pdp(pdp);
 	}
 }
 
-int apn_stop(struct apn_ctx *apn, bool force)
+int apn_stop(struct apn_ctx *apn)
 {
-	LOGPAPN(LOGL_NOTICE, apn, "%sStopping\n", force ? "FORCED " : "");
+	LOGPAPN(LOGL_NOTICE, apn, "Stopping\n");
 	/* check if pools have any active PDP contexts and bail out */
 	pool_close_all_pdp(apn->v4.pool);
 	pool_close_all_pdp(apn->v6.pool);
@@ -223,7 +204,7 @@ int apn_start(struct apn_ctx *apn)
 		LOGPAPN(LOGL_INFO, apn, "Opening Kernel GTP device %s\n", apn->tun.cfg.dev_name);
 		if (apn->cfg.apn_type_mask & (APN_TYPE_IPv6|APN_TYPE_IPv4v6)) {
 			LOGPAPN(LOGL_ERROR, apn, "Kernel GTP currently supports only IPv4\n");
-			apn_stop(apn, false);
+			apn_stop(apn);
 			return -1;
 		}
 		if (gsn == NULL) {
@@ -257,7 +238,7 @@ int apn_start(struct apn_ctx *apn)
 				apn->v4.cfg.ifconfig_prefix.prefixlen)) {
 			LOGPAPN(LOGL_ERROR, apn, "Failed to set tun IPv4 address %s: %s\n",
 				in46p_ntoa(&apn->v4.cfg.ifconfig_prefix), strerror(errno));
-			apn_stop(apn, false);
+			apn_stop(apn);
 			return -1;
 		}
 	}
@@ -270,7 +251,7 @@ int apn_start(struct apn_ctx *apn)
 			LOGPAPN(LOGL_ERROR, apn, "Failed to set tun IPv6 address %s: %s. "
 				"Ensure you have ipv6 support and not used the disable_ipv6 sysctl?\n",
 				in46p_ntoa(&apn->v6.cfg.ifconfig_prefix), strerror(errno));
-			apn_stop(apn, false);
+			apn_stop(apn);
 			return -1;
 		}
 	}
@@ -283,7 +264,7 @@ int apn_start(struct apn_ctx *apn)
 			LOGPAPN(LOGL_ERROR, apn, "Failed to set tun IPv6 link-local address %s: %s. "
 				"Ensure you have ipv6 support and not used the disable_ipv6 sysctl?\n",
 				in46p_ntoa(&apn->v6.cfg.ll_prefix), strerror(errno));
-			apn_stop(apn, false);
+			apn_stop(apn);
 			return -1;
 		}
 		apn->v6_lladdr = apn->v6.cfg.ll_prefix.addr.v6;
@@ -301,7 +282,7 @@ int apn_start(struct apn_ctx *apn)
 		if (rc < 1) {
 			LOGPAPN(LOGL_ERROR, apn, "Cannot obtain IPv6 link-local address of interface: %s\n",
 				rc ? strerror(errno) : "tun interface has no link-local IP assigned");
-			apn_stop(apn, false);
+			apn_stop(apn);
 			return -1;
 		}
 		apn->v6_lladdr = ipv6_tun_linklocal_ip.addr.v6;
@@ -318,7 +299,7 @@ int apn_start(struct apn_ctx *apn)
 				blacklist, blacklist_size)) {
 			LOGPAPN(LOGL_ERROR, apn, "Failed to create IPv4 pool\n");
 			talloc_free(blacklist);
-			apn_stop(apn, false);
+			apn_stop(apn);
 			return -1;
 		}
 		talloc_free(blacklist);
@@ -335,7 +316,7 @@ int apn_start(struct apn_ctx *apn)
 				blacklist, blacklist_size)) {
 			LOGPAPN(LOGL_ERROR, apn, "Failed to create IPv6 pool\n");
 			talloc_free(blacklist);
-			apn_stop(apn, false);
+			apn_stop(apn);
 			return -1;
 		}
 		talloc_free(blacklist);
@@ -365,7 +346,8 @@ static bool send_trap(const struct gsn_t *gsn, const struct pdp_t *pdp, const st
 static int delete_context(struct pdp_t *pdp)
 {
 	struct gsn_t *gsn = pdp->gsn;
-	struct apn_ctx *apn = pdp->priv;
+	struct pdp_priv_t *pdp_priv = pdp->priv;
+	struct apn_ctx *apn;
 	struct ippoolm_t *member;
 	int i;
 
@@ -376,212 +358,32 @@ static int delete_context(struct pdp_t *pdp)
 			member = pdp->peer[i];
 			send_trap(gsn, pdp, member, "imsi-rem-ip"); /* TRAP with IP removal */
 			ippool_freeip(member->pool, member);
-		} else if(i == 0)
+		} else if (i == 0) {
 			LOGPPDP(LOGL_ERROR, pdp, "Cannot find/free IP Pool member\n");
+		}
 	}
 
-	if (apn->cfg.gtpu_mode == APN_GTPU_MODE_KERNEL_GTP) {
+	if (!pdp_priv) {
+		LOGPPDP(LOGL_NOTICE, pdp, "Deleting PDP context: without private structure!\n");
+		return 0;
+	}
+
+	/* Remove from SGSN */
+	sgsn_peer_remove_pdp_priv(pdp_priv);
+
+	apn = pdp_priv->apn;
+	if (apn && apn->cfg.gtpu_mode == APN_GTPU_MODE_KERNEL_GTP) {
 		if (gtp_kernel_tunnel_del(pdp, apn->tun.cfg.dev_name)) {
 			LOGPPDP(LOGL_ERROR, pdp, "Cannot delete tunnel from kernel:%s\n",
 				strerror(errno));
 		}
 	}
 
-	return 0;
-}
-
-#include <osmocom/gsm/tlv.h>
-
-/* RFC 1332 */
-enum ipcp_options {
-	IPCP_OPT_IPADDR = 3,
-	IPCP_OPT_PRIMARY_DNS = 129,
-	IPCP_OPT_SECONDARY_DNS = 131,
-};
-
-struct ipcp_option_hdr {
-	uint8_t type;
-	uint8_t len;
-	uint8_t data[0];
-} __attribute__ ((packed));
-
-struct ipcp_hdr {
-	uint8_t code;
-	uint8_t id;
-	uint16_t len;
-	uint8_t options[0];
-} __attribute__ ((packed));
-
-/* determine if IPCP contains given option */
-static struct ipcp_option_hdr *ipcp_contains_option(struct ipcp_hdr *ipcp, enum ipcp_options opt)
-{
-	uint8_t *cur = ipcp->options;
-
-	/* iterate over Options and check if protocol contained */
-	while (cur + 2 <= ((uint8_t *)ipcp) + ntohs(ipcp->len)) {
-		struct ipcp_option_hdr *cur_opt = (struct ipcp_option_hdr *) cur;
-		if (cur_opt->type == opt)
-			return cur_opt;
-		cur += cur_opt->len;
-	}
-	return NULL;
-}
-
-/* 3GPP TS 24.008 10.6.5.3 */
-enum pco_protocols {
-	PCO_P_LCP		= 0xC021,
-	PCO_P_PAP		= 0xC023,
-	PCO_P_CHAP		= 0xC223,
-	PCO_P_IPCP		= 0x8021,
-	PCO_P_PCSCF_ADDR	= 0x0001,
-	PCO_P_IM_CN_SS_F	= 0x0002,
-	PCO_P_DNS_IPv6_ADDR	= 0x0003,
-	PCO_P_POLICY_CTRL_REJ	= 0x0004,	/* only in Network->MS */
-	PCO_P_MS_SUP_NETREQ_BCI	= 0x0005,
-	/* reserved */
-	PCO_P_DSMIPv6_HA_ADDR	= 0x0007,
-	PCO_P_DSMIPv6_HN_PREF	= 0x0008,
-	PCO_P_DSMIPv6_v4_HA_ADDR= 0x0009,
-	PCO_P_IP_ADDR_VIA_NAS	= 0x000a,	/* only MS->Network */
-	PCO_P_IPv4_ADDR_VIA_DHCP= 0x000b,	/* only MS->Netowrk */
-	PCO_P_PCSCF_IPv4_ADDR	= 0x000c,
-	PCO_P_DNS_IPv4_ADDR	= 0x000d,
-	PCO_P_MSISDN		= 0x000e,
-	PCO_P_IFOM_SUPPORT	= 0x000f,
-	PCO_P_IPv4_LINK_MTU	= 0x0010,
-	PCO_P_MS_SUPP_LOC_A_TFT	= 0x0011,
-	PCO_P_PCSCF_RESEL_SUP	= 0x0012,	/* only MS->Network */
-	PCO_P_NBIFOM_REQ	= 0x0013,
-	PCO_P_NBIFOM_MODE	= 0x0014,
-	PCO_P_NONIP_LINK_MTU	= 0x0015,
-	PCO_P_APN_RATE_CTRL_SUP	= 0x0016,
-	PCO_P_PS_DATA_OFF_UE	= 0x0017,
-	PCO_P_REL_DATA_SVC	= 0x0018,
-};
-
-/* determine if PCO contains given protocol */
-static uint8_t *pco_contains_proto(struct ul255_t *pco, uint16_t prot)
-{
-	uint8_t *cur = pco->v + 1;
-
-	/* iterate over PCO and check if protocol contained */
-	while (cur + 3 <= pco->v + pco->l) {
-		uint16_t cur_prot = osmo_load16be(cur);
-		uint8_t cur_len = cur[2];
-		if (cur_prot == prot)
-			return cur;
-		cur += cur_len + 3;
-	}
-	return NULL;
-}
-
-/*! Get the peer of pdp based on IP version used.
- *  \param[in] pdp PDP context to select the peer from.
- *  \param[in] v4v6 IP version to select. Valid values are 4 and 6.
- *  \returns The selected peer matching the given IP version. NULL if not present.
- */
-static struct ippoolm_t *pdp_get_peer_ipv(struct pdp_t *pdp, bool is_ipv6) {
-	uint8_t len1, len2, i;
-
-	if (is_ipv6) {
-		len1 = 8;
-		len2 = 16;
-	} else {
-		len1 = sizeof(struct in_addr);
-		len2 = len1;
-	}
-
-	for (i = 0; i < 2; i++) {
-		struct ippoolm_t * ippool = pdp->peer[i];
-		if (ippool && (ippool->addr.len == len1 || ippool->addr.len == len2))
-			return ippool;
-	}
-	return NULL;
-}
-
-/* construct an IPCP PCO response from request*/
-static int build_ipcp_pco(struct apn_ctx *apn, struct pdp_t *pdp, struct msgb *msg)
-{
-	const struct in46_addr *dns1 = &apn->v4.cfg.dns[0];
-	const struct in46_addr *dns2 = &apn->v4.cfg.dns[1];
-	struct ipcp_hdr *ipcp;
-	uint8_t *len1, *len2, *pco_ipcp;
-	uint8_t *start = msg->tail;
-	unsigned int len_appended;
-
-	if (!(pco_ipcp = pco_contains_proto(&pdp->pco_req, PCO_P_IPCP)))
-		return 0;
-	ipcp = (struct ipcp_hdr*) (pco_ipcp + 3);  /* 2=type + 1=len */
-
-	/* Three byte T16L header */
-	msgb_put_u16(msg, 0x8021);	/* IPCP */
-	len1 = msgb_put(msg, 1);	/* Length of contents: delay */
-
-	msgb_put_u8(msg, 0x02);		/* ACK */
-	msgb_put_u8(msg, ipcp->id);	/* ID: Needs to match request */
-	msgb_put_u8(msg, 0x00);		/* Length MSB */
-	len2 = msgb_put(msg, 1);	/* Length LSB: delay */
-
-	if (dns1->len == 4 && ipcp_contains_option(ipcp, IPCP_OPT_PRIMARY_DNS)) {
-		msgb_put_u8(msg, 0x81);		/* DNS1 Tag */
-		msgb_put_u8(msg, 2 + dns1->len);/* DNS1 Length, incl. TL */
-		msgb_put_u32(msg, ntohl(dns1->v4.s_addr));
-	}
-
-	if (dns2->len == 4 && ipcp_contains_option(ipcp, IPCP_OPT_SECONDARY_DNS)) {
-		msgb_put_u8(msg, 0x83);		/* DNS2 Tag */
-		msgb_put_u8(msg, 2 + dns2->len);/* DNS2 Length, incl. TL */
-		msgb_put_u32(msg, ntohl(dns2->v4.s_addr));
-	}
-
-	/* patch in length values */
-	len_appended = msg->tail - start;
-	*len1 = len_appended - 3;
-	*len2 = len_appended - 3;
+	talloc_free(pdp_priv);
 
 	return 0;
 }
 
-/* process one PCO request from a MS/UE, putting together the proper responses */
-static void process_pco(struct apn_ctx *apn, struct pdp_t *pdp)
-{
-	struct msgb *msg = msgb_alloc(256, "PCO");
-	struct ippoolm_t *peer_v4 = pdp_get_peer_ipv(pdp, false);
-	unsigned int i;
-
-	OSMO_ASSERT(msg);
-	msgb_put_u8(msg, 0x80); /* ext-bit + configuration protocol byte */
-
-	if (peer_v4)
-		build_ipcp_pco(apn, pdp, msg);
-
-	if (pco_contains_proto(&pdp->pco_req, PCO_P_DNS_IPv6_ADDR)) {
-		for (i = 0; i < ARRAY_SIZE(apn->v6.cfg.dns); i++) {
-			struct in46_addr *i46a = &apn->v6.cfg.dns[i];
-			if (i46a->len != 16)
-				continue;
-			msgb_t16lv_put(msg, PCO_P_DNS_IPv6_ADDR, i46a->len, i46a->v6.s6_addr);
-		}
-	}
-
-	if (pco_contains_proto(&pdp->pco_req, PCO_P_DNS_IPv4_ADDR)) {
-		for (i = 0; i < ARRAY_SIZE(apn->v4.cfg.dns); i++) {
-			struct in46_addr *i46a = &apn->v4.cfg.dns[i];
-			if (i46a->len != 4)
-				continue;
-			msgb_t16lv_put(msg, PCO_P_DNS_IPv4_ADDR, i46a->len, (uint8_t *)&i46a->v4);
-		}
-	}
-
-	if (msgb_length(msg) > 1) {
-		memcpy(pdp->pco_neg.v, msgb_data(msg), msgb_length(msg));
-		pdp->pco_neg.l = msgb_length(msg);
-	} else
-		pdp->pco_neg.l = 0;
-
-	msgb_free(msg);
-}
-
 static bool apn_supports_ipv4(const struct apn_ctx *apn)
 {
 	if (apn->v4.cfg.static_prefix.addr.len  || apn->v4.cfg.dynamic_prefix.addr.len)
@@ -596,6 +398,36 @@ static bool apn_supports_ipv6(const struct apn_ctx *apn)
 	return false;
 }
 
+static struct sgsn_peer* ggsn_find_sgsn(struct ggsn_ctx *ggsn, struct in_addr *peer_addr)
+{
+	struct sgsn_peer *sgsn;
+
+	llist_for_each_entry(sgsn, &ggsn->sgsn_list, entry) {
+		if (memcmp(&sgsn->addr, peer_addr, sizeof(*peer_addr)) == 0)
+			return sgsn;
+	}
+	return NULL;
+}
+
+static struct sgsn_peer* ggsn_find_or_create_sgsn(struct ggsn_ctx *ggsn, struct pdp_t *pdp)
+{
+	struct sgsn_peer *sgsn;
+	struct in_addr ia;
+
+	if (gsna2in_addr(&ia, &pdp->gsnrc)) {
+		LOGPPDP(LOGL_ERROR, pdp, "Failed parsing gsnrc (len=%u) to discover SGSN\n",
+			pdp->gsnrc.l);
+		return NULL;
+	}
+
+	if ((sgsn = ggsn_find_sgsn(ggsn, &ia)))
+		return sgsn;
+
+	sgsn = sgsn_peer_allocate(ggsn, &ia, pdp->version);
+	llist_add(&sgsn->entry, &ggsn->sgsn_list);
+	return sgsn;
+}
+
 int create_context_ind(struct pdp_t *pdp)
 {
 	static char name_buf[256];
@@ -604,15 +436,19 @@ int create_context_ind(struct pdp_t *pdp)
 	struct in46_addr addr[2];
 	struct ippoolm_t *member = NULL, *addrv4 = NULL, *addrv6 = NULL;
 	char straddrv4[INET_ADDRSTRLEN], straddrv6[INET6_ADDRSTRLEN];
-	struct apn_ctx *apn;
+	struct apn_ctx *apn = NULL;
 	int rc, num_addr, i;
+	char *apn_name;
+	struct sgsn_peer *sgsn;
+	struct pdp_priv_t *pdp_priv;
 
-	osmo_apn_to_str(name_buf, pdp->apn_req.v, pdp->apn_req.l);
-
-	LOGPPDP(LOGL_DEBUG, pdp, "Processing create PDP context request for APN '%s'\n", name_buf);
+	apn_name = osmo_apn_to_str(name_buf, pdp->apn_req.v, pdp->apn_req.l);
+	LOGPPDP(LOGL_DEBUG, pdp, "Processing create PDP context request for APN '%s'\n",
+		apn_name ? name_buf : "(NONE)");
 
 	/* First find an exact APN name match */
-	apn = ggsn_find_apn(ggsn, name_buf);
+	if (apn_name != NULL)
+		apn = ggsn_find_apn(ggsn, name_buf);
 	/* ignore if the APN has not been started */
 	if (apn && !apn->started)
 		apn = NULL;
@@ -648,9 +484,15 @@ int create_context_ind(struct pdp_t *pdp)
 		return 0;
 	}
 
+	/* Store the actual APN for logging and the VTY */
+	rc = osmo_apn_from_str(pdp->apn_use.v, sizeof(pdp->apn_use.v), apn->cfg.name);
+	if (rc < 0) /* Unlikely this would happen, but anyway... */
+		LOGPPDP(LOGL_ERROR, pdp, "Failed to store APN '%s'\n", apn->cfg.name);
+	pdp->apn_use.l = rc;
+
 	/* Allocate dynamic addresses from the pool */
 	for (i = 0; i < num_addr; i++) {
-		if (addr[i].len == sizeof(struct in_addr)) {
+		if (in46a_is_v4(&addr[i])) {
 			/* does this APN actually have an IPv4 pool? */
 			if (!apn_supports_ipv4(apn))
 				goto err_wrong_af;
@@ -663,7 +505,7 @@ int create_context_ind(struct pdp_t *pdp)
 
 			addrv4 = member;
 
-		} else if (addr[i].len == sizeof(struct in6_addr)) {
+		} else if (in46a_is_v6(&addr[i])) {
 
 			/* does this APN actually have an IPv6 pool? */
 			if (!apn_supports_ipv6(apn))
@@ -700,7 +542,14 @@ int create_context_ind(struct pdp_t *pdp)
 	}
 
 	pdp->ipif = apn->tun.tun;	/* TODO */
-	pdp->priv = apn;
+
+	pdp_priv = talloc_zero(ggsn, struct pdp_priv_t);
+	pdp->priv = pdp_priv;
+	pdp_priv->lib = pdp;
+	/* Create sgsn and assign pdp to it */
+	sgsn = ggsn_find_or_create_sgsn(ggsn, pdp);
+	sgsn_peer_add_pdp_priv(sgsn, pdp_priv);
+	pdp_priv->apn = apn;
 
 	/* TODO: change trap to send 2 IPs */
 	if (!send_trap(gsn, pdp, member, "imsi-ass-ip")) { /* TRAP with IP assignment */
@@ -764,7 +613,7 @@ static int cb_tun_ind(struct tun_t *tun, void *pack, unsigned len)
 		pool = apn->v6.pool;
 		break;
 	default:
-		LOGP(DTUN, LOGL_NOTICE, "non-IPv%u packet received from tun\n", iph->version);
+		LOGTUN(LOGL_NOTICE, tun, "non-IPv%u packet received\n", iph->version);
 		return -1;
 	}
 
@@ -772,26 +621,21 @@ static int cb_tun_ind(struct tun_t *tun, void *pack, unsigned len)
 	if (!pool)
 		return 0;
 
-	DEBUGP(DTUN, "Received packet for APN(%s) from tun %s", apn->cfg.name, tun->devname);
-
 	if (ippool_getip(pool, &ipm, &dst)) {
-		DEBUGPC(DTUN, " with no PDP contex! (%s)\n", iph->version == 4 ?
+		LOGTUN(LOGL_DEBUG, tun, "Received packet for APN(%s) with no PDP contex! (%s)\n",
+			apn->cfg.name,
+			iph->version == 4 ?
 			inet_ntop(AF_INET, &iph->saddr, straddr, sizeof(straddr)) :
 			inet_ntop(AF_INET6, &ip6h->ip6_src, straddr, sizeof(straddr)));
 		return 0;
 	}
-	DEBUGPC(DTUN, "\n");
+	LOGTUN(LOGL_DEBUG, tun, "Received packet for APN(%s)\n", apn->cfg.name);
 
 	if (ipm->peer)		/* Check if a peer protocol is defined */
 		gtp_data_req(apn->ggsn->gsn, (struct pdp_t *)ipm->peer, pack, len);
 	return 0;
 }
 
-/* RFC3307 link-local scope multicast address */
-static const struct in6_addr all_router_mcast_addr = {
-	.s6_addr = { 0xff,0x02,0,0,  0,0,0,0, 0,0,0,0,  0,0,0,2 }
-};
-
 /* MS-originated GTP1-U packet, needs to be sent via TUN device */
 static int encaps_tun(struct pdp_t *pdp, void *pack, unsigned len)
 {
@@ -855,8 +699,6 @@ static int encaps_tun(struct pdp_t *pdp, void *pack, unsigned len)
 	return tun_encaps((struct tun_t *)pdp->ipif, pack, len);
 }
 
-static char *config_file = "osmo-ggsn.cfg";
-
 /* callback for tun device osmocom select loop integration */
 static int ggsn_tun_fd_cb(struct osmo_fd *fd, unsigned int what)
 {
@@ -892,51 +734,53 @@ static int ggsn_gtp_fd_cb(struct osmo_fd *fd, unsigned int what)
 	return rc;
 }
 
-static void ggsn_gtp_tmr_start(struct ggsn_ctx *ggsn)
+/* libgtp callback for confirmations */
+static int cb_conf(int type, int cause, struct pdp_t *pdp, void *cbp)
 {
-	struct timeval next;
+	struct sgsn_peer *sgsn;
+	int rc = 0;
 
-	/* Retrieve next retransmission as timeval */
-	gtp_retranstimeout(ggsn->gsn, &next);
+	if (cause == EOF)
+		LOGP(DGGSN, LOGL_NOTICE, "libgtp EOF (type=%u, pdp=%p, cbp=%p)\n",
+			type, pdp, cbp);
 
-	/* re-schedule the timer */
-	osmo_timer_schedule(&ggsn->gtp_timer, next.tv_sec, next.tv_usec/1000);
-}
-
-/* timer callback for libgtp retransmission and ping */
-static void ggsn_gtp_tmr_cb(void *data)
-{
-	struct ggsn_ctx *ggsn = data;
-
-	/* do all the retransmissions as needed */
-	gtp_retrans(ggsn->gsn);
-
-	ggsn_gtp_tmr_start(ggsn);
-}
-
-/* To exit gracefully. Used with GCC compilation flag -pg and gprof */
-static void signal_handler(int s)
-{
-	LOGP(DGGSN, LOGL_NOTICE, "signal %d received\n", s);
-	switch (s) {
-	case SIGINT:
-	case SIGTERM:
-		LOGP(DGGSN, LOGL_NOTICE, "SIGINT received, shutting down\n");
-		end = 1;
+	switch (type) {
+	case GTP_DELETE_PDP_REQ:
+		/* Remark: We actually never reach this path nowadays because
+		   only place where we call gtp_delete_context_req2() is during
+		   ggsn_close_one_pdp() path, and in that case we free all pdp
+		   contexts immediatelly without waiting for confirmation
+		   (through gtp_freepdp_teardown()) since we want to tear down
+		   the whole APN anyways. As a result, DeleteCtxResponse will
+		   never reach here since it will be dropped at some point in
+		   lower layers in the Rx path. This code is nevertheless left
+		   here in order to ease future developent and avoid possible
+		   future memleaks once more scenarios where GGSN sends a
+		   DeleteCtxRequest are introduced. */
+		if (pdp)
+			rc = gtp_freepdp(pdp->gsn, pdp);
 		break;
-	case SIGABRT:
-	case SIGUSR1:
-		talloc_report(tall_vty_ctx, stderr);
-		talloc_report_full(tall_ggsn_ctx, stderr);
-		break;
-	case SIGUSR2:
-		talloc_report_full(tall_vty_ctx, stderr);
-		break;
-	default:
+	case GTP_ECHO_REQ:
+		sgsn = (struct sgsn_peer *)cbp;
+		sgsn_peer_echo_resp(sgsn, cause == EOF);
 		break;
 	}
+	return rc;
 }
 
+static int cb_recovery3(struct gsn_t *gsn, struct sockaddr_in *peer, struct pdp_t *pdp, uint8_t recovery)
+{
+	struct ggsn_ctx *ggsn = (struct ggsn_ctx *)gsn->priv;
+	struct sgsn_peer *sgsn;
+
+	sgsn = ggsn_find_sgsn(ggsn, &peer->sin_addr);
+	if (!sgsn) {
+		LOGPGGSN(LOGL_NOTICE, ggsn, "Received Recovery IE for unknown SGSN (no PDP contexts active)\n");
+		return -EINVAL;
+	}
+
+	return sgsn_peer_handle_recovery(sgsn, pdp, recovery);
+}
 
 /* Start a given GGSN */
 int ggsn_start(struct ggsn_ctx *ggsn)
@@ -977,12 +821,11 @@ int ggsn_start(struct ggsn_ctx *ggsn)
 	rc = osmo_fd_register(&ggsn->gtp_fd1u);
 	OSMO_ASSERT(rc == 0);
 
-	/* Start GTP re-transmission timer */
-	osmo_timer_setup(&ggsn->gtp_timer, ggsn_gtp_tmr_cb, ggsn);
-
 	gtp_set_cb_data_ind(ggsn->gsn, encaps_tun);
 	gtp_set_cb_delete_context(ggsn->gsn, delete_context);
 	gtp_set_cb_create_context_ind(ggsn->gsn, create_context_ind);
+	gtp_set_cb_conf(ggsn->gsn, cb_conf);
+	gtp_set_cb_recovery3(ggsn->gsn, cb_recovery3);
 
 	LOGPGGSN(LOGL_NOTICE, ggsn, "Successfully started\n");
 	ggsn->started = true;
@@ -1003,9 +846,7 @@ int ggsn_stop(struct ggsn_ctx *ggsn)
 
 	/* iterate over all APNs and stop them */
 	llist_for_each_entry(apn, &ggsn->apn_list, list)
-		apn_stop(apn, true);
-
-	osmo_timer_del(&ggsn->gtp_timer);
+		apn_stop(apn);
 
 	osmo_fd_unregister(&ggsn->gtp_fd1u);
 	osmo_fd_unregister(&ggsn->gtp_fd1c);
@@ -1019,127 +860,3 @@ int ggsn_stop(struct ggsn_ctx *ggsn)
 	ggsn->started = false;
 	return 0;
 }
-
-static void print_usage()
-{
-	printf("Usage: osmo-ggsn [-h] [-D] [-c configfile] [-V]\n");
-}
-
-static void print_help()
-{
-	printf(	"  Some useful help...\n"
-		"  -h --help		This help text\n"
-		"  -D --daemonize	Fork the process into a background daemon\n"
-		"  -c --config-file	filename The config file to use\n"
-		"  -V --version		Print the version of OsmoGGSN\n"
-		);
-}
-
-static void handle_options(int argc, char **argv)
-{
-	while (1) {
-		int option_index = 0, c;
-		static struct option long_options[] = {
-			{ "help", 0, 0, 'h' },
-			{ "daemonize", 0, 0, 'D' },
-			{ "config-file", 1, 0, 'c' },
-			{ "version", 0, 0, 'V' },
-			{ 0, 0, 0, 0 }
-		};
-
-		c = getopt_long(argc, argv, "hdc:V", long_options, &option_index);
-		if (c == -1)
-			break;
-
-		switch (c) {
-		case 'h':
-			print_usage();
-			print_help();
-			exit(0);
-		case 'D':
-			daemonize = 1;
-			break;
-		case 'c':
-			config_file = optarg;
-			break;
-		case 'V':
-			print_version(1);
-			exit(0);
-			break;
-		}
-	}
-}
-
-int main(int argc, char **argv)
-{
-	struct ggsn_ctx *ggsn;
-	int rc;
-
-	tall_ggsn_ctx = talloc_named_const(NULL, 0, "OsmoGGSN");
-	msgb_talloc_ctx_init(tall_ggsn_ctx, 0);
-	g_vty_info.tall_ctx = tall_ggsn_ctx;
-
-	/* Handle keyboard interrupt SIGINT */
-	signal(SIGINT, &signal_handler);
-	signal(SIGTERM, &signal_handler);
-	signal(SIGABRT, &signal_handler);
-	signal(SIGUSR1, &signal_handler);
-	signal(SIGUSR2, &signal_handler);
-
-	osmo_init_ignore_signals();
-	osmo_init_logging2(tall_ggsn_ctx, &log_info);
-	osmo_stats_init(tall_ggsn_ctx);
-
-	vty_init(&g_vty_info);
-	logging_vty_add_cmds(NULL);
-	osmo_talloc_vty_add_cmds();
-	osmo_stats_vty_add_cmds(&log_info);
-	ggsn_vty_init();
-	ctrl_vty_init(tall_ggsn_ctx);
-
-	handle_options(argc, argv);
-
-	rate_ctr_init(tall_ggsn_ctx);
-
-	rc = vty_read_config_file(config_file, NULL);
-	if (rc < 0) {
-		fprintf(stderr, "Failed to open config file: '%s'\n", config_file);
-		exit(2);
-	}
-
-	rc = telnet_init_dynif(tall_ggsn_ctx, NULL, vty_get_bind_addr(), OSMO_VTY_PORT_GGSN);
-	if (rc < 0)
-		exit(1);
-
-	g_ctrlh = ctrl_interface_setup(NULL, OSMO_CTRL_PORT_GGSN, NULL);
-	if (!g_ctrlh) {
-		LOGP(DGGSN, LOGL_ERROR, "Failed to create CTRL interface.\n");
-		exit(1);
-	}
-
-	if (daemonize) {
-		rc = osmo_daemonize();
-		if (rc < 0) {
-			perror("Error during daemonize");
-			exit(1);
-		}
-	}
-
-#if 0
-	/* qos                                                             */
-	qos.l = 3;
-	qos.v[2] = (args_info.qos_arg) & 0xff;
-	qos.v[1] = ((args_info.qos_arg) >> 8) & 0xff;
-	qos.v[0] = ((args_info.qos_arg) >> 16) & 0xff;
-#endif
-
-	/* Main select loop */
-	while (!end) {
-		osmo_select_main(0);
-	}
-
-	llist_for_each_entry(ggsn, &g_ggsn_list, list)
-		ggsn_stop(ggsn);
-
-	return 1;
-}
