@@ -6,6 +6,7 @@
 #include <osmocom/core/linuxlist.h>
 #include <osmocom/core/select.h>
 #include <osmocom/core/timer.h>
+#include <osmocom/ctrl/control_if.h>
 
 #include "../lib/tun.h"
 #include "../lib/ippool.h"
@@ -13,6 +14,8 @@
 #include "../lib/in46_addr.h"
 #include "../gtp/gtp.h"
 
+#include "sgsn.h"
+
 #define APN_TYPE_IPv4	0x01	/* v4-only */
 #define APN_TYPE_IPv6	0x02	/* v6-only */
 #define APN_TYPE_IPv4v6	0x04	/* v4v6 dual-stack */
@@ -88,6 +91,14 @@ struct apn_ctx {
 	struct apn_ctx_ip v6;
 };
 
+struct pdp_priv_t {
+	struct pdp_t *lib; /* pointer to libgtp associated pdp_t instance */
+	struct sgsn_peer *sgsn;
+	struct apn_ctx *apn;
+	struct llist_head entry; /* to be included into sgsn_peer */
+	/* struct ggsn_ctx can be reached through lib->gsn->priv, or through sgsn->ggsn */
+};
+
 struct ggsn_ctx {
 	/* global list of GGSNs */
 	struct llist_head list;
@@ -95,6 +106,9 @@ struct ggsn_ctx {
 	/* list of APNs in this GGSN */
 	struct llist_head apn_list;
 
+	/* list of SGSN peers (struct sgsn_peer) in this GGSN. TODO: hash table with key <ip+port>? */
+	struct llist_head sgsn_list;
+
 	bool started;
 
 	struct {
@@ -111,6 +125,8 @@ struct ggsn_ctx {
 		struct in46_addr gtpu_addr;
 		/* directory for state file */
 		char *state_dir;
+		/* Time between Echo requests on each SGSN */
+		unsigned int echo_interval;
 		/* administratively shut-down (true) or not (false) */
 		bool shutdown;
 	} cfg;
@@ -122,8 +138,6 @@ struct ggsn_ctx {
 	struct osmo_fd gtp_fd0;
 	struct osmo_fd gtp_fd1c;
 	struct osmo_fd gtp_fd1u;
-
-	struct osmo_timer_list gtp_timer;
 };
 
 /* ggsn_vty.c */
@@ -135,9 +149,24 @@ struct ggsn_ctx *ggsn_find_or_create(void *ctx, const char *name);
 struct apn_ctx *ggsn_find_apn(struct ggsn_ctx *ggsn, const char *name);
 struct apn_ctx *ggsn_find_or_create_apn(struct ggsn_ctx *ggsn, const char *name);
 
-/* ggsn.c */
+/* ggsn_main.c */
+extern struct ctrl_handle *g_ctrlh;
 extern void *tall_ggsn_ctx;
+
+/* ggsn.c */
 extern int ggsn_start(struct ggsn_ctx *ggsn);
 extern int ggsn_stop(struct ggsn_ctx *ggsn);
 extern int apn_start(struct apn_ctx *apn);
-extern int apn_stop(struct apn_ctx *apn, bool force);
+extern int apn_stop(struct apn_ctx *apn);
+void ggsn_close_one_pdp(struct pdp_t *pdp);
+
+#define LOGPAPN(level, apn, fmt, args...)			\
+	LOGP(DGGSN, level, "APN(%s): " fmt, (apn)->cfg.name, ## args)
+
+#define LOGPGGSN(level, ggsn, fmt, args...)			\
+	LOGP(DGGSN, level, "GGSN(%s): " fmt, (ggsn)->cfg.name, ## args)
+
+#define LOGPPDP(level, pdp, fmt, args...) LOGPDPX(DGGSN, level, pdp, fmt, ## args)
+
+#define LOGTUN(level, tun, fmt, args...) \
+	LOGP(DTUN, level, "TUN(%s): " fmt, (tun)->devname, ## args)
