@@ -26,16 +26,16 @@ int ippool_printaddr(struct ippool_t *this)
 {
 	unsigned int n;
 	printf("ippool_printaddr\n");
-	printf("Firstdyn %d\n", this->firstdyn - this->member);
-	printf("Lastdyn %d\n", this->lastdyn - this->member);
-	printf("Firststat %d\n", this->firststat - this->member);
-	printf("Laststat %d\n", this->laststat - this->member);
-	printf("Listsize %d\n", this->listsize);
+	printf("Firstdyn %td\n", this->firstdyn - this->member);
+	printf("Lastdyn %td\n", this->lastdyn - this->member);
+	printf("Firststat %td\n", this->firststat - this->member);
+	printf("Laststat %td\n", this->laststat - this->member);
+	printf("Listsize %u\n", this->listsize);
 
 	for (n = 0; n < this->listsize; n++) {
 		char s[256];
 		in46a_ntop(&this->member[n].addr, s, sizeof(s));
-		printf("Unit %d inuse %d prev %d next %d addr %s\n",
+		printf("Unit %d inuse %d prev %td next %td addr %s\n",
 		       n,
 		       this->member[n].inuse,
 		       this->member[n].prev - this->member,
@@ -202,7 +202,7 @@ int ippool_new(struct ippool_t **this, const struct in46_prefix *dyn, const stru
 	/* Parse only first instance of pool for now */
 
 	int i;
-	struct in46_addr addr;
+	struct in46_addr addr = { 0 };
 	size_t addrprefixlen;
 	struct in46_addr stataddr;
 	size_t stataddrprefixlen;
@@ -276,9 +276,8 @@ int ippool_new(struct ippool_t **this, const struct in46_prefix *dyn, const stru
 	(*this)->hashmask = (*this)->hashsize - 1;
 
 	/* Allocate hash table */
-	if (!
-	    ((*this)->hash =
-	     calloc(sizeof(struct ippoolm_t), (*this)->hashsize))) {
+	(*this)->hash = calloc((*this)->hashsize, sizeof(struct ippoolm_t *));
+	if (!(*this)->hash) {
 		SYS_ERR(DIP, LOGL_ERROR, 0,
 			"Failed to allocate memory for hash members in ippool");
 		return -1;
@@ -513,7 +512,15 @@ int ippool_newip(struct ippool_t *this, struct ippoolm_t **member,
 		p2->next = NULL;
 		p2->prev = NULL;
 		p2->inuse = 2;	/* Static address in use */
-		memcpy(&p2->addr, addr, sizeof(addr));
+		/* p2->addr.len and addr->len already match (see above). */
+		if (p2->addr.len == sizeof(struct in_addr))
+			p2->addr.v4 = addr->v4;
+		else if (p2->addr.len == sizeof(struct in6_addr))
+			p2->addr.v6 = addr->v6;
+		else {
+			SYS_ERR(DIP, LOGL_ERROR, 0, "MS requested unsupported PDP context type");
+			return -GTPCAUSE_UNKNOWN_PDP;
+		}
 		*member = p2;
 		(void)ippool_hashadd(this, *member);
 		if (0)
